// -*- C++ -*-
// $Id: FittedRegion3D.h,v 1.3 2005/04/11 12:47:38 heller Exp $
/* class FittedRegion3D
 *
 * This class adds boundary information as well as some useful output
 * methods to FittedPlane3D.
 *
 * A hull-finding algorithm is available that constructs proper instances
 * for the hull and holes members below. It relies on valid ij-coords
 * of the in- and outlier points.
 *
 * A convex hull algorithm is available but needs to be adapted.
 * Ask me.
 *
 * created  Mon Nov 20 14:41:06 1995  Justus Piater
 */ 
#ifndef FittedRegion3D_h
#define FittedRegion3D_h
#include <Plane3D.h>

class FittedRegion3D : public FittedPlane3D {
private:
  void init ();
#ifdef INVENTOR
  void output_inventor_normal (ofstream &os);
#endif

public:
  bool left_handed;		// XYZ-system left-handed?
  TokenList *hull;		// set of Point3D
  TokenSet *holes;		// set of TokenLists of Point3D
  
  FittedRegion3D ();
  FittedRegion3D (FittedPlane3D &);
  FittedRegion3D (FittedRegion3D &);
  virtual ~FittedRegion3D ();

  FittedRegion3D operator= (FittedRegion3D &);

  void trace();
  char *name() const { return "FittedRegion3D"; }
  void output (Isrostream& os);
  void input (Isristream& is);

#ifdef INVENTOR
  void output_inventor (ofstream &);
  void output_inventor (ofstream &, bool points, bool poles = true);
  void output_polygon (ofstream &os_xyz, ofstream &os_polygons,
		       bool raw = false);
#endif
};
#endif
