#include <Transform3D.h>
#include <string.h>		// memset

#ifdef __GNUC__
#define IGNORE __attribute__ ((unused))
#else
#define IGNORE
#endif

#if 0
static char rcsid[] IGNORE = "$Id: Transform3D.cc,v 1.3 1996/02/05 13:51:45 bdraper Exp $";

/* initialize global variables */
Transform2D *MIRROR_X = new Transform2D(-1.0, 0.0, 0.0, 0.0, 1.0, 0.0);
Transform2D *MIRROR_Y = new Transform2D(1.0, 0.0, 0.0, 0.0, -1.0, 0.0);
Transform2D *SWAP_XY = new Transform2D(0.0, 1.0, 0.0, 1.0, 0.0, 0.0);
CoordSys2D *UNIT_XWINDOW = new CoordSys2D;
CoordSys2D *UNIT_ROWCOL = new CoordSys2D( SWAP_XY, UNIT_XWINDOW);
CoordSys2D *UNIT_XY = new CoordSys2D(*(tr_translate(0.0, 1.0)) * (*MIRROR_Y), 
				     UNIT_XWINDOW);
CoordSys2D *UNIT_MID_LEFT = new CoordSys2D(*(tr_translate(0.5, 0.5)) * (*MIRROR_Y), 
					       UNIT_XWINDOW);
CoordSys2D *UNIT_MID_RIGHT = new CoordSys2D( MIRROR_X, UNIT_MID_LEFT );
#endif


Transform3D::Transform3D()
{
  memset (matrix, 0, sizeof (matrix));
  matrix[0][0] = matrix[1][1] = matrix[2][2] = matrix[3][3] = 1.0;
}

Transform3D::Transform3D(Transform3D* ts)
{
  for (int i=0;i < 4;i++) {
    for (int j=0;j < 4;j++) {
      matrix[i][j] = ts->matrix[i][j];
    }
  }
}

Transform3D::Transform3D(double r11, double r12, double r13, double tx,
			 double r21, double r22, double r23, double ty,
			 double r31, double r32, double r33, double tz,
			 double a0,  double a1,  double a2, double s)
{
  matrix[0][0] = r11;
  matrix[0][1] = r12;
  matrix[0][2] = r13;
  matrix[0][3] = tx;
  matrix[1][0] = r21;
  matrix[1][1] = r22;
  matrix[1][2] = r23;
  matrix[1][3] = ty;
  matrix[2][0] = r31;
  matrix[2][1] = r32;
  matrix[2][2] = r33;
  matrix[2][3] = tz;
  matrix[3][0] = a0;
  matrix[3][1] = a1;
  matrix[3][2] = a2;
  matrix[3][3] = s;
}

ostream &operator<< (ostream &os, const Transform3D &trafo)
{
  os << "Transform3D::\n";
  os << trafo.matrix[0][0] << " " << trafo.matrix[0][1] << " "
     << trafo.matrix[0][2] << " " << trafo.matrix[0][3] << "\n";
  os << trafo.matrix[1][0] << " " << trafo.matrix[1][1] << " "
     << trafo.matrix[1][2] << " " << trafo.matrix[1][3] << "\n";
  os << trafo.matrix[2][0] << " " << trafo.matrix[2][1] << " "
     << trafo.matrix[2][2] << " " << trafo.matrix[2][3] << "\n";
  os << trafo.matrix[3][0] << " " << trafo.matrix[3][1] << " "
     << trafo.matrix[3][2] << " " << trafo.matrix[3][3] << "\n";
  return os;
}

void Transform3D::output(Isrostream& os) 
{
  os << "Transform3D::\n";
  os << matrix[0][0] << " " << matrix[0][1] << " "
     << matrix[0][2] << " " << matrix[0][3] << "\n";
  os << matrix[1][0] << " " << matrix[1][1] << " "
     << matrix[1][2] << " " << matrix[1][3] << "\n";
  os << matrix[2][0] << " " << matrix[2][1] << " "
     << matrix[2][2] << " " << matrix[2][3] << "\n";
  os << matrix[3][0] << " " << matrix[3][1] << " "
     << matrix[3][2] << " " << matrix[3][3] << "\n";
}

void Transform3D::input(Isristream& is) 
{
  is >> "Transform3D::\n";
  is >> matrix[0][0] >> matrix[0][1] >> matrix[0][2] >> matrix[0][3];
  is >> matrix[1][0] >> matrix[1][1] >> matrix[1][2] >> matrix[1][3];
  is >> matrix[2][0] >> matrix[2][1] >> matrix[2][2] >> matrix[2][3];
  is >> matrix[3][0] >> matrix[3][1] >> matrix[3][2] >> matrix[3][3];
}

Transform3D* operator * (Transform3D& t1, Transform3D& t2)
{
  Transform3D *new_t = new Transform3D;

  for(int i=0; i < 4; i++)
    for(int j=0; j < 4; j++) {
      double sum = 0.0;
      for(int k=0; k < 4; k++) 
	sum += t1.matrix[j][k] * t2.matrix[k][i];
      new_t->matrix[j][i] = sum;
    }

  return new_t;
}

void Transform3D::Apply(double x, double y, double z, double s,
			double &tx, double &ty, double &tz, double &ts)
{
  tx = x*matrix[0][0] + y*matrix[0][1] + z*matrix[0][2] + s*matrix[0][3];
  ty = x*matrix[1][0] + y*matrix[1][1] + z*matrix[1][2] + s*matrix[1][3];
  tz = x*matrix[2][0] + y*matrix[2][1] + z*matrix[2][2] + s*matrix[2][3];
  ts = x*matrix[3][0] + y*matrix[3][1] + z*matrix[3][2] + s*matrix[3][3];

  if (ts != 0.0) {
    tx /= ts;
    ty /= ts;
    tz /= ts;
    ts = 1.0;
  }
}

void Transform3D::Apply4D(double x, double y, double z, double s,
			  double &tx, double &ty, double &tz, double &ts)
{
  tx = x*matrix[0][0] + y*matrix[0][1] + z*matrix[0][2] + s*matrix[0][3];
  ty = x*matrix[1][0] + y*matrix[1][1] + z*matrix[1][2] + s*matrix[1][3];
  tz = x*matrix[2][0] + y*matrix[2][1] + z*matrix[2][2] + s*matrix[2][3];
  ts = x*matrix[3][0] + y*matrix[3][1] + z*matrix[3][2] + s*matrix[3][3];
}
      
int Transform3D::Apply(double x, double y, double z,
		       double &tx, double &ty, double &tz)
{
  double s = 1.0;
  double ts;

  Apply(x, y, z, s, tx, ty, tz, ts);
  if (ts == 0.0) return 0;
  else return 1;
}


double Transform3D::Determinant()
{
#define COL(i,j) ((i)>=(j) ? (i)+1 : (i))
  double det = 0.0;
  for (int j = 0; j < 4; j++) {
    Transform2D A1j
      (matrix[1][COL(0,j)], matrix[1][COL(2,j)], matrix[1][COL(3,j)],
       matrix[2][COL(0,j)], matrix[2][COL(2,j)], matrix[2][COL(3,j)],
       matrix[1][COL(0,j)], matrix[1][COL(2,j)], matrix[1][COL(3,j)]);

    det += ((j%2) ? -1 : 1) * matrix[0][j] * A1j.Determinant();
  }
  return det;
}


Transform3D *Transform3D::Inverse()
// source: ftp://replicant.csci.unt.edu/pub/oleg/LinAlg.shar
//    (see ftp://replicant.csci.unt.edu/pub/oleg/README.html)
{
  Transform3D *inverse = new Transform3D (this);

  // from LinAlg.h:
  typedef double REAL;
  const int ncols = 4, nrows = 4;
  REAL *elements = (double *) inverse->matrix;

  // from matrix1.cc:
  REAL *index[ncols];
  REAL * col_p;
  int i;
  for(i=0,col_p=&elements[0]; i<ncols; i++,col_p += nrows)
    index[i] = col_p;

  /* The following code has been copied from matrix_inv.cc with hardly
     any changes: */
/*
 *------------------------------------------------------------------------
 *
 *		The most general (Gauss-Jordan) matrix inverse
 *
 * This method works for any matrix (which of course must be square and
 * non-singular). Use this method only if none of specialized algorithms
 * (for symmetric, tridiagonal, etc) matrices isn't applicable/available.
 * Also, the matrix to invert has to be _well_ conditioned:
 * Gauss-Jordan eliminations (even with pivoting) perform poorly for
 * near-singular matrices (e.g., Hilbert matrices).
 *
 * The method inverts matrix inplace and returns the determinant if
 * determ_ptr was specified as not nil. determinant will be exactly zero
 * if the matrix turns out to be (numerically) singular. If determ_ptr is
 * nil and matrix happens to be singular, throw up.
 *
 * The algorithm perform inplace Gauss-Jordan eliminations with
 * full pivoting. It was adapted from my Algol-68 "translation" (ca 1986)
 * of the FORTRAN code described in
 * Johnson, K. Jeffrey, "Numerical methods in chemistry", New York,
 * N.Y.: Dekker, c1980, 503 pp, p.221
 *
 * Note, since it's much more efficient to perform operations on matrix
 * columns rather than matrix rows (due to the layout of elements in the
 * matrix), the present method implements a "transposed" algorithm.
 *
 */

  double determinant = 1;
  const double singularity_tolerance = 1e-35;

				// Locations of pivots (indices start with 0)
  struct Pivot { short row, col; } pivots[ncols];
  char was_pivoted[nrows]; memset(was_pivoted,0,sizeof(was_pivoted));

  Pivot *pivotp;
  for(pivotp = &pivots[0]; pivotp < &pivots[ncols]; pivotp++)
  {
    short prow = 0, pcol = 0;		// Location of a pivot to be
    {					// Look through all non-pivoted cols
      REAL max_value = 0;		// (and rows) for a pivot (max elem)
      for(register int j=0; j<ncols; j++)
	if( !was_pivoted[j] )
	{
	  register REAL * cp;
	  register int k;
	  REAL curr_value = 0;
	  for(k=0,cp=index[j]; k<nrows; k++,cp++)
	    if( !was_pivoted[k] && (curr_value = fabs(*cp)) > max_value )
	      max_value = curr_value, prow = k, pcol = j;
	}
      if( max_value < singularity_tolerance ) {
	cerr << "Matrix turns out to be singular: can't invert";
	exit (-1);
      }
      pivotp->row = prow;
      pivotp->col = pcol;
    }

    if( prow != pcol )			// Swap prow-th and pcol-th columns to
    {					// bring the pivot to the diagonal
      register REAL * cr = index[prow];
      register REAL * cc = index[pcol];
      for(register int k=0; k<nrows; k++)
      {
	REAL temp = *cr; *cr++ = *cc; *cc++ = temp;
      }
    }
    was_pivoted[prow] = 1;

    {					// Normalize the pivot column and
      register REAL * pivot_cp = index[prow];
      double pivot_val = pivot_cp[prow];	// pivot is at the diagonal
      determinant *= pivot_val;		// correct the determinant
      pivot_cp[prow] = 1;
      for(register int k=0; k<nrows; k++)
	*pivot_cp++ /= pivot_val;
    }

    {					// Perform eliminations
      register REAL * pivot_rp = elements + prow;	// pivot row
      for(register int k=0; k<nrows; k++, pivot_rp += nrows)
	if( k != prow )
	{
	  double temp = *pivot_rp;
	  *pivot_rp = 0;
	  register REAL * pivot_cp = index[prow];	// pivot column
	  register REAL * elim_cp  = index[k];		// elimination column
	  for(register int l=0; l<nrows; l++)
	    *elim_cp++ -= temp * *pivot_cp++;
	}
    }
  }

  int no_swaps = 0;		// Swap exchanged *rows* back in place
  for(pivotp = &pivots[ncols-1]; pivotp >= &pivots[0]; pivotp--)
    if( pivotp->row != pivotp->col )
    {
      no_swaps++;
      register REAL * rp = elements + pivotp->row;
      register REAL * cp = elements + pivotp->col;
      for(register int k=0; k<ncols; k++, rp += nrows, cp += nrows)
      {
	REAL temp = *rp; *rp = *cp; *cp = temp;
      }
    }

  return inverse;
}

int Transform3D::operator== (const Transform3D& other)
{
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++)	{
      if (fabs(matrix[i][j] - other.matrix[i][j]) > FUZZ) return 0;
    }
  }
  return 1;
}

Transform3D *tr_translate(double x, double y, double z)
{
  return new Transform3D(1.0, 0.0, 0.0, x,
			 0.0, 1.0, 0.0, y,
			 0.0, 0.0, 1.0, z);
}

Transform3D *tr_scale3D(double mag_factor)
{
  return new Transform3D(mag_factor, 0.0, 0.0, 0.0,
			 0.0, mag_factor, 0.0, 0.0,
			 0.0, 0.0, mag_factor, 0.0);
}

Transform3D *tr_scale3D(double xscale, double yscale, double zscale)
{
  return new Transform3D(xscale, 0.0, 0.0, 0.0,
			 0.0, yscale, 0.0, 0.0,
			 0.0, 0.0, zscale, 0.0);
}

Transform3D *tr_rotate (double amount, Axis axis, RotationUnit measure)
{
  if (measure == Degrees) amount *= (M_PI / 180.0);
  switch (axis) {
  case xAxis: return tr_rotate_x (sin (amount), cos (amount));
  case yAxis: return tr_rotate_y (sin (amount), cos (amount));
  case zAxis: return tr_rotate_z (sin (amount), cos (amount));
  }
  return NULL;			// cannot be reached
}

Transform3D *tr_rotate_x (double s, double c)
{
  return new Transform3D (1.0, 0.0, 0.0, 0.0,
			  0.0, c, -s, 0.0,
			  0.0, s, c, 0.0);
}

Transform3D *tr_rotate_y (double s, double c)
{
  return new Transform3D (c, 0.0, s, 0.0,
			  0.0, 1.0, 0.0, 0.0,
			  -s, 0.0, c, 0.0);
}

Transform3D *tr_rotate_z (double s, double c)
{
  return new Transform3D (c, -s, 0.0, 0.0,
			  s, c, 0.0, 0.0,
			  0.0, 0.0, 1.0, 0.0);
}


CoordSys3D::CoordSys3D()
{
  sys2base = NULL;
  base2sys = new Transform3D();
  set_sys2base();
}

CoordSys3D::CoordSys3D(Transform3D *trans, CoordSys3D *base)
{
  sys2base = NULL;
  base2sys = (*trans) * (*base->base2sys);
  set_sys2base();
}

CoordSys3D::CoordSys3D(CoordSys3D *sys)
{
  sys2base = NULL;
  base2sys = new Transform3D(sys->base2sys);
  set_sys2base();
}

void CoordSys3D::set_sys2base()
{
  delete sys2base;
  sys2base = base2sys->Inverse();
  if (base2sys == NULL) {
    cout << "ERROR: Coodinate system transforms must be invertible.\n";
    //exit(-1);
    return;
  }
}

void CoordSys3D::output(Isrostream& os)
{
  os << "CoordSys3D:: base2sys = " << base2sys << "\n";
  os << "   sys2base = " << sys2base << "\n";
}
  
void CoordSys3D::input(Isristream& is)
{
  is >> "CoordSys3D:: base2sys = ";
  base2sys = (Transform3D *)read_token_ptr(is);
  is >> "sys2base =";
  sys2base = (Transform3D *)read_token_ptr(is);
}

void CoordSys3D::trace()
{
  Register(base2sys);
  Register(sys2base);
}

Transform3D *overlay(CoordSys3D* source, CoordSys3D* dest)
{
  /* overlay returns the identity transform if one or both coordinate
     systems are NULL */
  if ((source == NULL) || (dest == NULL)) return new Transform3D;
  else return (*dest->base2sys) * (*source->sys2base);
}


CoordSys3D *CoordSys3D::Translate(double DX, double DY, double DZ)
{
  Transform3D *trans = tr_translate(DX,DY,DZ);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}

CoordSys3D *CoordSys3D::Scale(double SC)
{
  Transform3D *trans = tr_scale3D(SC);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}

CoordSys3D *CoordSys3D::Scale(double xscale, double yscale, double zscale)
{
  Transform3D *trans = tr_scale3D(xscale, yscale, zscale);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}

CoordSys3D *CoordSys3D::Rotate_x(double amount, RotationUnit measure)
{
  Transform3D* trans = tr_rotate_x(amount, measure);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}

CoordSys3D *CoordSys3D::Rotate_y(double amount, RotationUnit measure)
{
  Transform3D* trans = tr_rotate_y(amount, measure);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}

CoordSys3D *CoordSys3D::Rotate_z(double amount, RotationUnit measure)
{
  Transform3D* trans = tr_rotate_z(amount, measure);
  CoordSys3D *new_sys = new CoordSys3D(trans, this);
  delete trans;
  return new_sys;
}
