// -*- C++ -*-
#ifndef TRANSFORM3D_DEFINED
#define TRANSFORM3D_DEFINED
#include<Token.h>
#include <Transform2D.h>	// RotationUnit
#include<math.h>
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef BOOLEAN
#define BOOLEAN int
#endif

enum Axis { xAxis, yAxis, zAxis };

class Transform3D : public Token {
  friend ostream &operator<< (ostream &os, const Transform3D &trafo);
  friend Transform3D* operator* (Transform3D& t1, Transform3D& t2);
 public:
  Transform3D();   /* returns identity tranform */
  Transform3D(double r11, double r12, double r13, double tx,
	      double r21, double r22, double r23, double ty,
	      double r31, double r32, double r33, double tz,
	      double a0 = 0.0,  double a1 = 0.0,  double a2 = 0.0,
	      double s = 1.0);
  Transform3D(Transform3D* ts);

  void output(Isrostream&);
  void input(Isristream&);
  char *name() const {return "Transform3D";}

  double Determinant();
//  double Minor(int, int);
  Transform3D *Inverse();
  void Apply(double x, double y, double z, double s,
	     double &tx, double &ty, double &tz, double &ts);
  void Apply4D(double x, double y, double z, double s,
	       double &tx, double &ty, double &tz, double &ts);
  int Apply(double x, double y, double z, double &tx, double &ty, double &tz);
  int operator== (const Transform3D& other);
  inline int operator!= (const Transform3D& other)
	{return (!operator== (other));}
  inline double element(int ix,int iy) {return matrix[ix][iy];}
 private:
  double matrix[4][4];
  const static double FUZZ = .00001;
};

Transform3D* operator* (Transform3D& t1, Transform3D& t2);

class CoordSys3D : public Token {
  friend Transform3D* overlay(CoordSys3D* source, CoordSys3D* dest);
 public:
  /* create base (identity) coordinate system */
  CoordSys3D();                     
  /* define new coordinate system in terms of a transform that converts
     it to any known base coordinate system */
  CoordSys3D(Transform3D *trans, CoordSys3D *base); 
  CoordSys3D(CoordSys3D *sys);
  ~CoordSys3D() { delete sys2base; delete base2sys; }

  char *name() {return "CoordSys3D";}
  void input(Isristream& is);
  void output(Isrostream& os);
  void trace();

  inline int operator== (const CoordSys3D& other)
  	{return *sys2base == *(other.sys2base);}
  inline int operator!= (const CoordSys3D& other)
	{return (!operator== (other));}
  CoordSys3D* Translate(double DX, double DY, double DZ);
  CoordSys3D* Scale(double SC);
  CoordSys3D* Scale(double xscale, double yscale, double zscale);
  CoordSys3D* Rotate_x(double amount, RotationUnit measure);
  CoordSys3D* Rotate_y(double amount, RotationUnit measure);
  CoordSys3D* Rotate_z(double amount, RotationUnit measure);
    
 private:
  Transform3D *sys2base;
  Transform3D *base2sys; 
  void set_sys2base();
};


Transform3D* overlay(CoordSys3D* source, CoordSys3D* dest);

/* HANDY DEFAULT TRANSFORMS & FUNCTIONS*/
#if 0
extern Transform2D *MIRROR_X;
extern Transform2D *MIRROR_Y;
extern Transform2D *MIRROR_Z;
extern Transform2D *SWAP_XY;
#endif
Transform3D *tr_translate(double x, double y, double z);
Transform3D *tr_scale3D(double mag_factor);
Transform3D *tr_scale3D(double xscale, double yscale, double zscale);
Transform3D *tr_rotate (double amount, Axis axis, RotationUnit measure);
Transform3D *tr_rotate_x (double sina, double cosa);
Transform3D *tr_rotate_y (double sina, double cosa);
Transform3D *tr_rotate_z (double sina, double cosa);

#if 0
/* HANDY DEFAULT COORDINATE SYSTEMS */
extern CoordSys2D *UNIT_XWINDOW;
extern CoordSys2D *UNIT_XY;
extern CoordSys2D *UNIT_ROWCOL;
extern CoordSys2D *UNIT_MID_RIGHT;
extern CoordSys2D *UNIT_MID_LEFT;
#endif

#endif /*  TRANSFORM3D_DEFINED */

