#include <Vector3D.h>

/* Vector3D.cc
   Created 10/23/95 by James Doyle
   This module provides the Vector3D class, which represents a vector
in 3D space y its i,j,k components.
*/

Vector3D::Vector3D() : i(0), j(0), k(0) {}

Vector3D::Vector3D(double newi, double newj, double newk) 
: i(newi), j(newj), k(newk) {}

void Vector3D::output(Isrostream& os)
{
  os << "Vector3D of (" << i << "," << j << "," << k << ")\n";
}

void Vector3D::input(Isristream& is)
{
  is >> "Vector3D of (" >> i >> "," >> j >> "," >> k >> ")";
}

bool Vector3D::operator==(const Vector3D& that) const
{
  return ((i==that.i) && (j==that.j) && (k==that.k));
}

Vector3D Vector3D::operator-(const Vector3D& that) const
{
  return Vector3D((i-that.i), (j-that.j), (k-that.k));
}

Vector3D Vector3D::cross(const Vector3D& that) const
{
  return Vector3D(
		  (j*that.k - k*that.j),
		  (k*that.i - i*that.k),
		  (i*that.j - j*that.i)
		  );
}

double Vector3D::dot(const Vector3D& that) const
{
  return (i*that.i)+(j*that.j)+(k*that.k);
}

Vector3D Vector3D::operator/(double scalar_divisor) const
{
  return Vector3D((i/scalar_divisor), (j/scalar_divisor), (k/scalar_divisor));
}

Vector3D Vector3D::operator*(double scalar_multiplier) const
{
  return Vector3D((i*scalar_multiplier), 
		  (j*scalar_multiplier), 
		  (k*scalar_multiplier));
}

bool Vector3D::is_unit() const
{
  return (dot(*this) == 1.0);
}



