#ifndef VECTOR3D_DEFINED
#define VECTOR3D_DEFINED
#include <Token.h>

/* Vector3D.h
   Created 10/23/95 by James Doyle
   This module provides the Vector3D class, which represents a vector
in 3D space by its i,j,k components.  Input and output, as well as
several mathematical methods, are provided.  The extra methods include:
equality, subtraction, scalar multiplication and division, cross
and dot products, and a check to return whether the vector is a unit
vector.
*/

class Vector3D : public Token {
  
 public:
  double i, j, k;
  
  Vector3D();
  Vector3D(double newi, double newj, double newk);
  Vector3D(const Vector3D& that) {i = that.i;j = that.j;k = that.k;}
  char *name() const { return "Vector3D"; }
  bool operator==(const Vector3D& that) const;
  Vector3D operator-(const Vector3D& that) const;
  Vector3D cross(const Vector3D& that) const;
  double dot(const Vector3D& that) const;
  Vector3D operator/(double scalar_divisor) const;
  Vector3D operator*(double scalar_multiplier) const;
  bool is_unit() const;

  void output(Isrostream& os);
  void input(Isristream& is);
};

#endif
